/*
 *
 *  Copyright (C) 2016 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#if defined(__PYTHON__)
  #include <Python.h>
#endif

#include <string.h>
#include <sys/stat.h>

#include "KCanLogFile.h"

// Aus einem log-File, welches durch can_logger erzeugt wurde, einen Array von
// KCanFrame erstellen.

KCanLogFile::KCanLogFile()
{
  Frames = NULL;
  count = 0;
}

KCanLogFile::~KCanLogFile()
{
  Init();
}

void KCanLogFile::Init()
{
  if (Frames)
    delete [] Frames;
  Frames = NULL;
}

unsigned KCanLogFile::LogFileToCanFrames(const KStream & stream, KCanFrame * Frames)
{
  char buf[100];
  KCanFrame frame;
  unsigned long pos = 0;
  unsigned count = 0;
  
  while (pos < stream.Size())
  {
    while (pos < stream.Size() && stream.GetByte(pos) < ' ')
      pos++;
    
    unsigned long off = pos + 1;
    while (off < stream.Size() && stream.GetByte(off) >= ' ')
      off++;
    if (off - pos < sizeof(buf))
    {
      strncpy(buf, stream.GetCharMemory() + pos, off - pos);
      buf[off - pos] = 0;
      if (frame.InitFromLog(buf))
      {
        if (Frames != NULL)
          Frames[count] = frame;
        count++;
      }
    }
    pos = off;
  }
  return count;
}

KCanFrame * KCanLogFile::BuildCanFrames(const KStream & stream)
{
  Init();
  
  count = LogFileToCanFrames(stream, NULL);
  if (count)
  {
    Frames = new KCanFrame[count];
    LogFileToCanFrames(stream, Frames);
  }
  
  return Frames;
}

bool KCanLogFile::ReadFile(const char * Filename)
{
  KStream stream;
  
  if (!stream.ReadFile(Filename))
  {
    printf("\nlog file \"%s\" not loaded\n", Filename);
    return false;
  }  
  
  printf("\nreading log file \"%s\" ...\n", Filename);
  BuildCanFrames(stream);
  printf("log file read: frame count %d\n\n", count);
  
  return Frames;
}


