/*
 *
 *  Copyright (C) 2015 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include "UBruecke.h"

#include <stdio.h>
#include <string.h>

#if defined(__BORLANDC__) && (__BORLANDC__ >= 0x640)
  #include <vcl.h>
#endif

#include "NUtils.h"
#include "KTcpCanClient.h"
#include "KElsterTable.h"
#include "KCanElster.h"
#include "KScanTable.h"
#include "KCanCommDriver.h"

volatile bool Stop = false;
t_dev dev = dev_simulation;
unsigned short Kennwort = 0;
KCanElster * CanElster = NULL;
#if defined(VCL)
  TComm * CS_Com = NULL; // TComm bietet RS232-Monitor an
  TComm * Null_CS = NULL;
#else
  KComm * CS_Com = NULL;
  KComm * Null_CS = NULL;
#endif
KCanElster * USBtin_Simul = NULL;

KTcpCanClient CanClient;
FILE * logFile = NULL;

const bool UseTimeStamp = true;

unsigned char not_connected_dev[256];


const char * copy_tstr(const my_char * from)
{
#if !defined(__BORLANDC__) || (__BORLANDC__ < 0x640) // XE2
  return from;
#else
  static char copy_str_char[1024];

  int i = 0;
  while (from[i])
  {
    copy_str_char[i] = from[i];
    i++;
  }
  copy_str_char[i] = 0;

  return copy_str_char;
#endif
}

bool GetElsterValue(unsigned short receiver_id, unsigned short elster_index, unsigned short & value)
{
  KComfortFrame send;
  KComfortFrame recv;
  KCanFrame Frame;

  Frame.InitElsterFrame(0x680, receiver_id, elster_index);
  Frame.Len = 7;
  send.GetCanFrame(Frame);
  bool Ok = Bruecke::ElsterComm(send, recv);
  if (Ok)
    Ok = recv.SetCanFrame(Frame);

  if (Ok)
    value = Frame.Value();

  return Ok;
}

namespace Bruecke
{

bool LoadScanTable(const char * Filename)
{
  bool Ok = KScanTable::LoadScanTable(Filename);
  printf("Tabelle \"%s\"", Filename);
  if (!Ok)
    printf(" nicht");
  printf(" geladen\n");

  return Ok;
}

void set_log_file(bool set, const char * Filename)
{
  if (set)
  {
    if (!logFile)
    {
      logFile = fopen(Filename, "ab+");
      if (!logFile)
        printf("Log-File \"%s\" konnte nicht geoeffnet werden.\n", Filename);
    }
  } else
  if (logFile)
  {
    fclose(logFile);
    logFile = NULL;
  }
}

void log_frame(bool send, const KComfortFrame & frame)
{
  if (logFile)
  {
    char str[62];

    strcpy(str, send ? "send:" : "recv:");
    for (int i = 0; i < 12; i++)
      sprintf(str + strlen(str), " %2.2x", frame.Data[i]);
    fprintf(logFile, "%s\n", str);
  }
}

bool ConnectedIndex(const KComfortFrame & cs, unsigned char & Index)
{
  KCanFrame Frame;

  bool Ok = cs.SetCanFrame(Frame);
  if (Ok)
    Ok = (Frame.Data[0] & 0x0f) == 1 &&
         Frame.GetElsterIdx() == 0x000b;

  Index = KScanTable::ShortCanId(Frame.ReceiverId());

  return Ok;
}

bool NotConnected(const KComfortFrame & cs)
{
  unsigned char Index;
  bool Ok = ConnectedIndex(cs, Index);

  if (Ok &&
      not_connected_dev[Index] > 1)
    return true;

  return false;
}

unsigned short GetKennwort()
{
  int i = 0;

  while (KComfortFrame::scan_geraete_id[i].elster_idx)
  {
    unsigned short value;
    unsigned short recv_id = KComfortFrame::scan_geraete_id[i].recv_id;
    unsigned short elster_idx = KComfortFrame::scan_geraete_id[i].elster_idx;

    bool Ok = GetElsterValue(recv_id, elster_idx, value) &&
              value != 0x8000 && value != 0;

    // Vor allem neuere Geräte, wie die WPM3, verwenden 0x0199 und 0x019a
    // als Versionsnummern. Der Wert an der Stelle 0x0199 ist meist grösser
    // als 256 und passt damit nicht in das alte Schema.
    if (Ok && elster_idx == 0x0199)
    {
      value = 0x8000;
    }
    if (Ok && value)
    {
      char str[64];

      SetValueType(str, et_dev_id, value);
      printf("Kennwort: %d  Version: %s\n", value, str);

      return value;
    }
    i++;
  }
  return 0;
}

bool InitKennwort(const char * Text)
{
  char str[32];
  bool shown = false;

  Kennwort = 0;
  TInt64 d = 0;

  if (dev == dev_nullmodem)
    return false;

  NUtils::GetInt(Text, d);
  if (0 < d && d <= 0xffff)
  {
    Kennwort = (unsigned short) d;
  } else
  if (dev == dev_can_server || dev == dev_can232 || dev == dev_simulation)
  {
    Kennwort = GetKennwort();
    shown = true;
    if (!Kennwort)
    {
      printf("ungueltiges Kennwort\n");
      return false;
    }
  }

  if (!shown)
  {
    SetValueType(str, et_dev_id, Kennwort);
    printf("Kennwort: %d  Version: %s\n", Kennwort, str);
  }
  return true;
}

static void LogElster(const KCanFrame & Frame)
{
  char str[64];
  char buf[128];

  if (FormElsterTable(Frame, str))
  {
    sprintf(buf, "Elster %3.3x.%4.4x.%4.4x: %s\n",
            Frame.Id, Frame.GetElsterIdx(), Frame.Value(), str);
    printf("%s", buf);
    if (logFile)
      fprintf(logFile, "%s", buf);
  }
}

static void LogElster(const KComfortFrame & cf)
{
  KCanFrame Frame;

  if (cf.SetCanFrame(Frame))
  {
    LogElster(Frame);
  }
}

static void LogSend(const KComfortFrame & cs_send)
{
  char str[64];

  cs_send.SetToString(str);
  printf("send CS: %s\n", str);
  if (logFile)
    fprintf(logFile, "send CS: %s\n", str);
}

static void LogRecv(const KComfortFrame & cs_recv)
{
  char str[64];

  cs_recv.SetToString(str);
  printf("recv CS: %s\n", str);
  if (logFile)
    fprintf(logFile, "recv CS: %s\n", str);
}

void WriteToCS(KComfortFrame * cs_recv, bool log)
{
  if (dev != dev_nullmodem)
  #if defined(VCL)
    CS_Com->WriteBufferToCOM(cs_recv->Data, 12);
  #else
    CS_Com->WriteBufferToCOM((const char *) (cs_recv->Data), 12);
  #endif
  if (log)
  {
    LogRecv(*cs_recv);
  }
}

#if defined(VCL)
  void Init_Simple_Com(TComm * & Com)
#else
  void Init_Simple_Com(KComm * & Com)
#endif
{
#if defined(VCL)
  Com = new TComm((TComponent *) NULL);
  Com->UseTimeStamp = UseTimeStamp;
#else
  Com = new KComm;
#endif
  Com->SetDTR(true);
  Com->SetRTS(true);
  Com->SetParity(2);
  Com->SetStopBits(2);
  Com->SetBaudrate(2400);
  Com->SetRtsFlowControl(false);
  Com->NewScreenSize(90, 2000);
  Com->SetTimerInterval(10);
  HideError = true;
}

void Init_CS_Com()
{
  Init_Simple_Com(CS_Com);
}

void Init_Null_Com()
{
  Init_Simple_Com(Null_CS);
}

bool ElsterComm(const KComfortFrame & send, KComfortFrame & recv)
{
  bool Ok = false;
#if defined(VCL)
  try
#endif
  {
    switch (dev)
    {
      case dev_nullmodem:
        return true;

      case dev_can_server:
        Ok = CanClient.SendComfortFrame(send, recv);
        break;

      case dev_can232:
        if (!send.SetCanFrame(CanElster->SendFrame) ||
            !CanElster->Send(1, true))
          return false;
        return recv.GetCanFrame(CanElster->RecvFrame);

      case dev_simulation:
        Ok = KScanTable::Get_scan_value(send, recv);
        if (Ok && send.GetMsgType() == KComfortFrame::mt_change)
          recv.SetOk(true);
        NUtils::SleepMs(Ok ? 50 : 100);
        break;

      default:
        break;
    }
  }
#if defined(VCL)
  catch (Exception &exception)
  {
    Application->ShowException(&exception);
    Ok = false;
    if (dev == dev_can_server)
      CanClient.Disconnect();
    CS_Com->ExitCOM();
  }
#endif
  return Ok;
}

bool InitCom(t_dev new_dev, int sec_com_nr, int timeout_ms,
             const my_char * remote_ip, bool binary_protocol,
             void * Form)
{
  bool Ok = true;

  dev = new_dev;

  if (dev == dev_can232)
  {
    char port[64];
    CanElster = new KCanElster;

    sprintf(port, "COM%d", sec_com_nr);
    if (!CanElster->Init(port))
      return false;

    CanElster->Trace = true;
    CanElster->Resume();

    return true;
  }

  if (dev == dev_can_server)
  {
    CanClient.mReadTimeout = timeout_ms;
    CanClient.SetRemoteHost(copy_tstr(remote_ip));
    CanClient.SetRemotePort(5524);
    CanClient.UseBinaryProtocol = binary_protocol;
    CanClient.mTrace = true;
    CanClient.IsServer = false;

    return true;
  }

  if (dev == dev_nullmodem)
  {
    if (!Null_CS)
    {
      Init_Null_Com();
    }
  #if defined(VCL)
    try
  #endif
    {
      Null_CS->SetComPortNr(sec_com_nr);
      Ok = Null_CS->InitCOM();
    }
  #if defined(VCL)
    catch (Exception &exception)
    {
      Application->ShowException(&exception);
      Ok = false;
    }
  #endif
  }

  if (dev == dev_simulation)
    return true;

  return Ok;
}

void CloseCom()
{
  CS_Com->ExitCOM();
  if (CanElster)
  {
    CanElster->Halt();
  #if defined(VCL)
    delete CanElster;
  #endif
    CanElster = NULL;
  }
  if (Null_CS)
    Null_CS->ExitCOM();
  if (USBtin_Simul)
  {
    USBtin_Simul->Halt();
    NUtils::SleepMs(1000);
    delete USBtin_Simul;
    USBtin_Simul = NULL;
  }
}

bool Start(int com_nr)
{
  Stop = false;
  if (com_nr > 0)
    CS_Com->SetComPortNr(com_nr);
#if defined(VCL)
  try
#endif
  {
    CS_Com->InitCOM();
  }
#if defined(VCL)
  catch (Exception &exception)
  {
    Application->ShowException(&exception);
  }
#endif
  if (!CS_Com->COM_INITIALISIERT())
  {
    CS_Com->Show();
    CS_Com->write(" serielle Schnittstelle konnte nicht geoeffnet werden!  ");
    CloseCom();
    return false;
  }

  return true;
}

void Execute(bool & cbxUseElsterTable, bool & cbxRecvTelegram,
             bool & cbxCanTelegram, bool & cbNotChange,
             bool & cbxCS_Telegram, bool & cbxRecvCS)
{
  if (dev == dev_nullmodem)
  {
    int recv_count = 0;
    int send_count = 0;
    KComfortFrame cs_recv;
    KComfortFrame cs_send;
    int cs_recv_data = 0;
    int cs_send_data = 0;
    do
    {
      int res;
      bool first = true;
      while ((res = CS_Com->ReadFromCOM()) != NULL_CHAR)
      {
        if (res < 0x100)
        {
          if (UseTimeStamp && first)
            CS_Com->WriteNow();
          Null_CS->WriteToCOM(res);
          cs_send.Data[cs_send_data++] = (unsigned short) res;
          if (cs_send_data >= 12)
          {
            if (cbxCS_Telegram)
            {
              LogSend(cs_send);
            }
            cs_send_data = 0;
          }
        } else
        if ((res & 0xffff) == 0x10)
        {
          CS_Com->SendBreak(20);
          Null_CS->SendBreak(20);
        }
        first = false;
      }

      first = true;
      while ((res = Null_CS->ReadFromCOM()) != NULL_CHAR)
      {
        if (res < 0x100)
        {
          if (UseTimeStamp && first)
            Null_CS->WriteNow();
          CS_Com->WriteToCOM(res);
          cs_recv.Data[cs_recv_data++] = (unsigned short) res;
          if (cs_recv_data >= 12)
          {
            if (cbxUseElsterTable)
            {
              LogElster(cs_recv);
            }
            if (cbxRecvCS)
            {
              LogRecv(cs_recv);
            }
            cs_recv_data = 0;
          }
        } else
        if ((res & 0xffff) == 0x10)
        {
          CS_Com->SendBreak(20);
        }
        recv_count = 0;
        first = false;
      }
      recv_count++;
      if (recv_count > 40)
      {
        recv_count = cs_recv_data = 0;
      }
      send_count++;
      if (send_count > 40)
      {
        send_count = cs_recv_data = 0;
      }
      NUtils::SleepMs(1);
    #if defined(VCL) || (defined(__BORLANDC__) && (__BORLANDC__ >= 0x640))
      Application->ProcessMessages();
    #endif
    } while (CS_Com->COM_INITIALISIERT() && !Stop);
    printf("== stop\n");
    CloseCom();

    return;
  }

  while (CS_Com->ReadFromCOM() != NULL_CHAR)   // Puffer leeren
    ;

  KComfortFrame cs;
  KComfortFrame cs_recv;
  bool written = false;
  int len_cs = 0;
  int len_cs_recv = 0;
  do
  {
  #if defined(__USBTIN__)
    if (USBtin_Simul)
      USBtin_Simul->ServeUSBtin();
  #endif    
    CS_Com->CommMonitor(NULL);
    if (dev == dev_nullmodem)
    {
      Null_CS->CommMonitor(NULL);
      int res2 = Null_CS->ReadFromCOM();
      if (res2 < 0x100)
      {
        CS_Com->WriteToCOM((char)(res2));
        cs_recv.Data[len_cs_recv++] = (unsigned char)res2;
      } else
      if (res2 > 0x100)
        len_cs_recv = 0;

      if (len_cs_recv == 12)
      {
        KCanFrame Frame;
        char str[64];

        log_frame(false, cs_recv);

        if (cbxUseElsterTable)
        {
          LogElster(cs_recv);
        }
        if (cbxRecvTelegram)
        {
          cs_recv.SetToString(str);
          printf("recv: %s\n", str);
        }
        len_cs_recv = 0;
      }
    }
    int res = CS_Com->ReadFromCOM();
    if (res < 0x100)
    {
      if (dev == dev_nullmodem)
        Null_CS->WriteToCOM((char)(res));

      cs.Data[len_cs++] = char(res);
      if (cs.Data[0] != 0x0d)
        len_cs = 0;

      if (len_cs == 12)
      {
        KCanFrame SendFrame;
        char str[64];
        KComfortFrame::msg_type cf_type;

        log_frame(true, cs);

        if (dev == dev_nullmodem)
          len_cs_recv = 0;

        // cs ==> SendFrame und cf_type
        if (!cs.SetCanFrame(SendFrame))
          cf_type = KComfortFrame::mt_invalid;
        else
          cf_type = cs.GetMsgType();
        if (cf_type == KComfortFrame::mt_short_change)
          cf_type = KComfortFrame::mt_change;

        // show send frame
        if (cf_type != KComfortFrame::mt_invalid)
        {
          if (cbxCanTelegram ||
              (cf_type == KComfortFrame::mt_change && !cbNotChange))
          {
            SendFrame.FormatFrame(false, str);
            printf("send: %s", str);
            if (logFile)
              fprintf(logFile, "send: %s", str);
          } else
          if (cbxCS_Telegram)
          {
            LogSend(cs);
          }
        }

        if ((cf_type == KComfortFrame::mt_change && cbNotChange))
        {
          LogElster(SendFrame);
        } else
        if (cf_type == KComfortFrame::mt_invalid)
        {
          cs.SetToString(str);
          printf("invalid: %s\n", str);
          cs_recv.SetOk(false);
          WriteToCS(&cs_recv, cbxRecvCS);
        } else
        if (dev == dev_nullmodem)
        {
          // do nothing
        } else
        if (cf_type == KComfortFrame::mt_init)
        {
          cs_recv.SetOkValue(Kennwort);
          // 03e8 005a
         /* cs_recv.SetOkValue(0x3e8);
          cs_recv.Data[6] = 0xa1;
          cs_recv.Data[7] = 0x99;
          KCanFrame f;
          f.InitElsterFrame(0x680, 0x680, 0x199);
          f.SetValue(0x3e8);
          f.Data[0]++;
          cs_recv.GetCanFrame(f);
          cs_recv.SetCheckSum();    */
          WriteToCS(&cs_recv, cbxRecvCS);
        } else
        if (cf_type == KComfortFrame::mt_get &&
            NotConnected(cs))
        {
          cs_recv.SetOk(false);
          WriteToCS(&cs_recv, cbxRecvCS);
        } else
        if (ElsterComm(cs, cs_recv)) // send and receive / interpret
        {
          log_frame(false, cs_recv);

          char buf[128];
          KCanFrame Frame;
          bool show_recv = false;
          bool recv_ok = cs_recv.SetCanFrame(Frame);
          if (cbxUseElsterTable &&
              cf_type == KComfortFrame::mt_change)
          {
            FormElsterTable(SendFrame, str);
            sprintf(buf, "Elster (sent %s) %3.3x.%4.4x.%4.4x: %s\n",
                    cs_recv.Data[0] == 0x55 ? "ok" : "failed",
                    SendFrame.ReceiverId(), SendFrame.GetElsterIdx(), SendFrame.Value(), str);
            printf("%s", buf);
            if (logFile)
              fprintf(logFile, "%s", buf);
          } else
          if (cbxUseElsterTable)
          {
            if (recv_ok)
            {
              LogElster(cs_recv);
            } else
            if (cs_recv.Data[0] != 0xaa && cs_recv.Data[0] != 0x55 &&
                !cbxRecvTelegram && cf_type != KComfortFrame::mt_init)
              show_recv = true;
          }
          if (cbxRecvTelegram || cf_type == KComfortFrame::mt_init || show_recv)
          {
            if (recv_ok &&cf_type != KComfortFrame::mt_init)
              Frame.FormatFrame(false, str);
            else {
              cs_recv.SetToString(str);
              strcat(str, "\n");
            }
            if (cf_type == KComfortFrame::mt_init)
              printf("init: %s", str);
            else
              printf("recv: %s", str);
          }
          unsigned char con_ind;
          if (ConnectedIndex(cs, con_ind))
          {
            if (cs_recv.Data[0] == 0xaa)
              not_connected_dev[con_ind]++;
          }
          WriteToCS(&cs_recv, cbxRecvCS);
        } // ElsterComm
        len_cs = 0;
        written = true;
      }
      if (written)
      {
      #if defined(VCL) || (defined(__BORLANDC__) && (__BORLANDC__ >= 0x640))
        Application->ProcessMessages();
      #endif
        if (dev != dev_nullmodem)
          CS_Com->writeln();
        written = false;
      }
    } else
    if (res != NULL_CHAR)
    {
      len_cs = 0;
      CS_Com->WriteChar(13);
      CS_Com->WriteChar(10);
    } else
      NUtils::SleepMs(1);
  #if defined(VCL) || (defined(__BORLANDC__) && (__BORLANDC__ >= 0x640))
    Application->ProcessMessages();
  #endif
  } while (CS_Com->COM_INITIALISIERT() && !Stop);
  printf("== stop\n");
  CloseCom();
}

void ExitProg()
{
#if defined(VCL)
  if (CanElster)
    delete CanElster;
  if (CS_Com)
    delete CS_Com;
#endif
  if (Null_CS)
    delete Null_CS;
}

void Send(const KCanFrame & Frame)
{
  if (USBtin_Simul)
    USBtin_Simul->SendData(Frame);
}

}

////////////////////////////////////////////////////////////////////////////////

#if defined(__NO_VCL__)

int main(int argc, char* argv[])
{
#if defined(__WINDOWS__)
  printf("cs_bruecke (simulation only)\n"
         "copyright (c) 2015 Juerg Mueller, CH-5524\n\n");
#else
  printf("cs_bruecke (simulation only)\n"
         "copyright (c) 2015 Jürg Müller, CH-5524\n\n");
#endif
  
  bool trace = !strcmp(argv[argc-1], "trace");
  if (trace)
    argc--;
  
  if (argc < 2 || argc > 3)
  {
    printf("usage:\n\n"
  #if defined(__WINDOWS__)
           "  cs_bruecke_dos <serial port to ComfortSoft> [ <scan_data table> ]\n\n"
           "  cs_bruecke_dos COM4 \"..\\scan_data.inc\"\n\n");
  #else
           "  ./cs_bruecke <serial port to ComfortSoft> [ <scan_data table> ]\n\n"
           "  ./cs_bruecke ttyABC ../scan_data.inc\n\n");
  #endif
    return 1;
  }

  if (argc >= 3 &&
      !Bruecke::LoadScanTable(argv[2]))
    return -1;

 // Bruecke::set_log_file(true, "log.txt");

  if (dev != dev_nullmodem &&
      !Bruecke::InitKennwort(""))
  {
    Bruecke::CloseCom();
    return -2;
  }

  Bruecke::Init_CS_Com();
  if (!Bruecke::InitCom(dev_simulation,
                        1, /* COM1 / second port not used */
                        500, /* timeout_ms */
                        "", false))
    return -3;

#if defined(__USBTIN__) // nicht ausgetestet!
  USBtin_Simul = new KCanElster;
  if (!USBtin_Simul->InitUSBtin("com6", true))
    return -5;
#endif

  CS_Com->SetComDev(argv[1]);
  if (!Bruecke::Start(0))
    return -4;

  printf("\n");

  bool flgUseElsterTable = true;
  bool flgRecvTelegram   = false;
  bool flgCanTelegram    = false;
  bool flgNotChange      = false;
  bool flgCS_Telegram    = false;
  bool flgRecvCS         = false;

  Bruecke::Execute(flgUseElsterTable,
                   flgRecvTelegram,
                   flgCanTelegram,
                   flgNotChange,
                   flgCS_Telegram,
                   flgRecvCS);

  return 0;
}

#endif

