/*
 *
 *  Copyright (C) 2016 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#include <stdio.h>
#include <string.h>

#include "NUtils.h"
#include "NCanUtils.h"
#include "KCanServer.h"
#include "KCanCommDriver.h"
#include "KCanElster.h"
#include "KElsterTable.h"

int main(int argc, char* argv[])
{
  printf("CAN to serial\n"
        "copyright (c) 2014 Jürg Müller, CH-5524\n\n");
  
  bool trace = !strcmp(argv[argc-1], "trace");
  if (trace)
    argc--;
  
  KCanElster Can;
  unsigned short sender_id = 0x680;
   
  bool cs_com = !strcmp(argv[argc-1], "cs_com");
  if (cs_com)
  {
    argc--;
    Can.driver_type = KCanServer::dt_cs;
  }
  Can.SetTrace(trace);
  
  if (argc == 4)
  {
    const char * ptr = argv[3];
    if (NUtils::IsHexDigit(*ptr))
      argc--;
    KComfortFrame::InitId = (unsigned short) NUtils::GetHex(ptr);
  }
  
  if (argc != 3)
  {
    printf("\nusage:\n"
           "  can2ser <can dev> <serial dev> [ <hex. CS-ID> ] [ \"cs_com\"] [ \"trace\" ]\n\n"
           "example: ./can2ser can0 ttys002\n\n");
    return -1;
  }
 
  KCanCS CanCS;
  if (!CanCS.Init(argv[2]))
  {
    printf("seral device \"%s\" not connected\n", argv[2]);
    Can.Halt();
    
    return -1;
  }  
 
  if (!Can.Init(argv[1]))
  {
    printf("device \"%s\" not connected\n", Can.GetDev());
    Can.Halt();
    
    return -1;
  }
  
  CanCS.Trace = trace;
  CanCS.AsSimulation = true;
  CanCS.SenderIsCS = true;
  Can.Trace = trace;
  Can.Resume();
  Can.SendCanId = sender_id;
    
  if (Can.GetInitCS_Id())
  {
    if (trace)
      printf("CS-Id: %4.4x\n", (int)KComfortFrame::InitId);
  }
  
  NUtils::SleepMs(1000);
  while (!Can.Terminated())
  {
    KCanFrame Frame;
    KComfortFrame cf;
    if (Can.GetFrame(Frame) &&
        cf.GetCanFrame(Frame))
    {
      if (trace)
        Frame.PrintFrame();
      CanCS.SendData(cf);
    }
    if (CanCS.ReceiveData(cf) &&
        cf.SetCanFrame(Frame))
    {
      NUtils::Time(Frame.TimeStampDay, Frame.TimeStampMs);
   
      if (cf.IsInitFrame())
      {
        cf.SetInitOk();
        CanCS.SendData(cf);
      } else
      // Vorsichtsmassnahmen:
      // - nur reine Abfragen zulassen
      // - Abfragen ohne 0xfa an das Gerät mit ID = 000 kann Nebenwirkungen
      //   auslösen! (deshalb nicht zulassen)
      if (cf.Data[3] == 0x01 && (cf.Data[4] & 0xf0) == 0 && // reine Abfragen
          (cf.Data[5] == 0xfa || cf.Data[2] != 0))
      {
        if (Can.SendData(Frame))
        {
          if (trace)
            Frame.PrintFrame();
        }
      } else
      if (true &&
         // Wert ändern
         // 0d 00 03-00-00 fa ...  (marc2012)
         cf.Data[3] == 0x00 && (cf.Data[4] & 0xf0) == 0 &&
         cf.Data[5] == 0xfa && cf.Data[2] != 0)
      {
        unsigned char u = Frame.Data[0] & 0xf0;
        Frame.Data[0] = u + 2;
        Can.SendFrame = Frame;
        bool Ok = Can.SetValue();
        cf.SetOk(Ok);
        if (!trace)
        {
          char str[128];
            
          printf("set param (%s)> ", Ok ? "true" : "false");
          if (FormElsterTable(Can.SendFrame, str))
          {
            printf("%s\n", str);
          }
          Frame.PrintFrame();
          CanCS.SendData(cf);
        }
      } else {
        if (!trace)
        {
          printf("no answer> ");
          Frame.PrintFrame();
        }
        cf.SetOk(false);
        CanCS.SendData(cf);
      }
    }
    NUtils::SleepMs(1);
  }
  
  Can.Halt(); 
  
  return 0;
}
